/* Copyright (C) |Meso|Star> 2015-2018 (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s3d_device_c.h"
#include "s3d_geometry.h"
#include "s3d_instance.h"
#include "s3d_mesh.h"
#include "s3d_scene_view_c.h"
#include "s3d_sphere.h"

#include <rsys/mem_allocator.h>

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static FINLINE void
sphere_ray_setup
  (RTCRay& ray,
   const float tfar,
   struct geometry* geom)
{
  const RTCRay r = ray;
  float cos_theta;
  ASSERT(tfar >= 0 && geom);

  ray.tfar = tfar;
  ray.geomID = geom->irtc;
  ray.primID = 0;
  f3_mulf(ray.Ng, ray.dir, tfar);
  f3_add(ray.Ng, ray.Ng, ray.org);
  f3_sub(ray.Ng, ray.Ng, geom->data.sphere->pos);

  /* Compute the parametric coordinate */
  f3_normalize(ray.Ng, ray.Ng);
  cos_theta = ray.Ng[2];
  ray.v = (1.f - cos_theta) * 0.5f;
  if(absf(cos_theta) == 1) {
    ray.u = 0;
  } else if(eq_epsf(ray.Ng[0], 0.f, 1.e-6f)) {
    ray.u = ray.Ng[1] > 0 ?  0.25f : 0.75f;
  } else {
    double phi = atan2f(ray.Ng[1],  ray.Ng[0]); /* phi in [-PI, PI] */
    if(phi < 0) phi = 2*PI + phi; /* phi in [0, 2PI] */
    ray.u = (float)(phi / (2*PI));
  }

  /* Filter the intersection if required */
  if(geom->data.sphere->filter.func) {
    rtc_hit_filter_wrapper(&geom->data.sphere->filter, ray);
    if(ray.geomID == RTC_INVALID_GEOMETRY_ID) ray = r; /* Restore the ray */
  }
}

static void
geometry_release(ref_T* ref)
{
  struct geometry* geom;
  struct s3d_device* dev;

  geom = CONTAINER_OF(ref, struct geometry, ref);
  dev = geom->dev;
  switch(geom->type) {
    case GEOM_MESH:
      if(geom->data.mesh) mesh_ref_put(geom->data.mesh);
      break;
    case GEOM_INSTANCE:
      if(geom->data.instance) instance_ref_put(geom->data.instance);
      break;
    case GEOM_SPHERE:
      if(geom->data.sphere) sphere_ref_put(geom->data.sphere);
      break;
    default: FATAL("Unreachable code\n"); break;
  }
  MEM_RM(dev->allocator, geom);
  S3D(device_ref_put(dev));
}

/*******************************************************************************
 * Local functions
 ******************************************************************************/
res_T
geometry_create
  (struct s3d_device* dev,
   struct geometry** out_geom)
{
  struct geometry* geom = NULL;
  res_T res = RES_OK;
  ASSERT(dev && out_geom);

  geom = (struct geometry*)MEM_CALLOC
    (dev->allocator, 1, sizeof(struct geometry));
  if(!geom) {
    res = RES_MEM_ERR;
    goto error;
  }
  ref_init(&geom->ref);
  S3D(device_ref_get(dev));
  geom->dev = dev;
  geom->name = S3D_INVALID_ID;
  geom->irtc = RTC_INVALID_GEOMETRY_ID;
  geom->flip_surface = 0;
  geom->is_enabled = 1;
  geom->type = GEOM_NONE;
  geom->data.mesh = NULL;

exit:
  *out_geom = geom;
  return res;
error:
  if(geom) {
    geometry_ref_put(geom);
    geom = NULL;
  }
  goto exit;
}

void
geometry_ref_get(struct geometry* geom)
{
  ASSERT(geom);
  ref_get(&geom->ref);
}

void
geometry_ref_put(struct geometry* geom)
{
  ASSERT(geom);
  ref_put(&geom->ref, geometry_release);
}

void
geometry_rtc_sphere_bounds(void* data, size_t item, RTCBounds& bounds)
{
  struct geometry* geom = (struct geometry*)data;
  struct sphere sphere;
  ASSERT(geom && item == 0 && geom->type == GEOM_SPHERE);
  (void)item;
  sphere = *geom->data.sphere;
  bounds.lower_x = sphere.pos[0] - sphere.radius;
  bounds.lower_y = sphere.pos[1] - sphere.radius;
  bounds.lower_z = sphere.pos[2] - sphere.radius;
  bounds.upper_x = sphere.pos[0] + sphere.radius;
  bounds.upper_y = sphere.pos[1] + sphere.radius;
  bounds.upper_z = sphere.pos[2] + sphere.radius;
}

void
geometry_rtc_sphere_intersect(void* data, RTCRay& ray, size_t item)
{
  float v[3];
  float A, B, C, D, Q, rcpA, t0, t1;
  struct geometry* geom = (struct geometry*)data;
  struct sphere sphere;
  ASSERT(geom && item == 0 && geom->type == GEOM_SPHERE);
  (void)item;

  sphere = *geom->data.sphere;
  f3_sub(v, ray.org, sphere.pos);
  A = f3_dot(ray.dir, ray.dir);
  B = 2*f3_dot(v, ray.dir);
  C = f3_dot(v, v) - sphere.radius*sphere.radius;
  D = B*B - 4*A*C;

  if(D < 0.0f) return;
  Q = (float)sqrt(D);
  rcpA = 1.f / A;
  t0 = 0.5f * rcpA * (-B - Q);
  t1 = 0.5f * rcpA * (-B + Q);

  if(ray.tnear < t0 && t0 < ray.tfar) sphere_ray_setup(ray, t0, geom);
  if(ray.tnear < t1 && t1 < ray.tfar) sphere_ray_setup(ray, t1, geom);
}

void
geometry_rtc_sphere_occluded(void* data, RTCRay& ray, size_t item)
{
  float v[3];
  float A, B, C, D, Q, rcpA, t0, t1;
  struct geometry* geom = (struct geometry*)data;
  struct sphere sphere;
  ASSERT(geom && item == 0 && geom->type == GEOM_SPHERE);
  (void)item;

  sphere = *geom->data.sphere;
  f3_sub(v, ray.org, sphere.pos);
  A = f3_dot(ray.dir, ray.dir);
  B = 2*f3_dot(v, ray.dir);
  C = f3_dot(v, v) - sphere.radius*sphere.radius;
  D = B*B - 4*A*C;

  if(D < 0.0f) return;
  Q = (float)sqrt(D);
  rcpA = 1.f / A;
  t0 = 0.5f * rcpA * (-B - Q);
  t1 = 0.5f * rcpA * (-B + Q);

  if(ray.tnear < t0 && t0 < ray.tfar) ray.geomID = 0;
  if(ray.tnear < t1 && t1 < ray.tfar) ray.geomID = 0;
}

